


#define	OPEN_RW		(OPEN_READ | OPEN_WRITE)
#define	OPEN		(OPEN_RW | OPEN_CREAT)

#define	REMOVE		(UNLINK | RMDIR | RENAME_TO | RENAME_FROM)

#define	CHMOD_CHOWN	(CHMOD | CHOWN | LCHOWN)


/* race protection entry is created for each file operation and inserted into
   a hash table
*/

struct rp_entry
{
	struct rp_entry	*next;
	int		tick;		/* keep track of time in terms of
					   ticks (hz / sec) */

	int		inode;
	int		fd;		/* file descriptor */
	int		fsid;		/* filesystem id */
	pid_t		pid;
	char		*name;		/* pathname */
	enum FILEOP	operation;
};


/* rp_purge() type defines */

#define	PURGE_TYPE_TIMEOUT	1	/* purge all stale entries */
#define	PURGE_TYPE_PID		2	/* purge entries with a certain pid */
#define	PURGE_TYPE_FD		4	/* purge entries with a certain fd */


/*
 * timeout values
 *
 * Suspicious directory/file operations are suspended for 10/3 + loadavg seconds
 * Directory operations are suspender for longer periods of time to account
 * for longer directory processing times
 */

#define	TIMEOUT_DIR		1000
#define	TIMEOUT_FILE		300


/* soft timer callout every second */
#define	TIMEOUT_TIMER		100

/* purge old entries every X seconds */
#define	TIMEOUT_PURGE		200
#define	TIMEOUT_PURGE_MAX	500


/* maximum number of free'd entries to cache */
#define FREE_LIST_NUM   500


/* rp_insert() operation types */
#define	RP_OP_NORMAL		1	/* normal case, insert entries into
					   hash */
#define	RP_OP_STAT		2	/* open(O_CREAT) case, insert stat,
					   lstat & access entries in a separate
					   hash */
#define	RP_OP_UNLINK		3	/* insert unlink'ed entries in a
					   separate hash */


/* rp_process() return codes */
#define	RP_PROCESS_OK		0	/* allow the system call to proceed */
#define	RP_PROCESS_INSERT	1	/* operation entry needs to be inserted */
#define	RP_PROCESS_RESPOND	2	/* suspend / kill process */
#define	RP_PROCESS_FAIL_CALL	3	/* current system call should be aborted */


/* deny/permit_all_policy() return codes */
#define	POLICY_OK		0
#define	POLICY_NOT_OK		1


/* OS independent function declarations */

int			deny_all_policy(int, int);
int			permit_all_policy(int, int);
int			rp_process(pid_t, int, int, char*, enum FILEOP, int);
int			rp_register_fd(pid_t, int, int);
int			rp_insert(pid_t, int, int, char*, enum FILEOP, int,int);
int			rp_insert_nolock(pid_t, int, int, char*, enum FILEOP, int,int);
void			rp_purge(int, int, int);
void			rp_timeout(void*);
void			rp_duplicate_entry(int, int);
void			rp_init();
void			rp_fini();
inline struct rp_entry*	rp_malloc();
inline void		rp_free();



/* Kernel module defines */


/* system call entry for a backup_systent table */

#define	BACKUP_SYSENT(num, e)	{ num, "sys_" ## #e, 0, raceprot_sys_##e },


/* call saved system call through a function pointer saved in backup_systent */

#define	CALL_SAVED_SYS_CALL(e)	backup_sysent[NUM_SYS_##e].old_sys_call(p, v, retval)


/* save all the information about mediated system calls into the following
   structure
 */

struct backup_sysent
{
	int		sys_number;		/* system call number */
	const char	*sys_name;		/* system call fn name */
	sy_call_t	*old_sys_call;		/* saved system call fn ptr */
	sy_call_t	*new_sys_call;		/* new system call fn ptr */
};


#define	BACKUP_SYSENT_NUM_ENTRIES	(sizeof(backup_sysent) / sizeof(struct backup_sysent))


extern int sys_lkmnosys();
extern int lkmexists __P((struct lkm_table *));


void raceprot_respond(struct proc*, enum FILEOP, const char*, int);
char* fileop_name(enum FILEOP operation);


#endif	/* __RACEPROT_H__ */
