/* DO NOT EDIT. AUTOMATICALLY GENERATED FILE. SEE README & generate.pl */

/*
 * Copyright (c) 2002-2004 Eugene Tsyrklevich (eugene@securityarchitects.com)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/mount.h>
#include <sys/exec.h>
#include <sys/lkm.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/syscall.h>
#include <sys/syscallargs.h>
#include <sys/proc.h>
#include <sys/namei.h>
#include <sys/vnode.h>
#include <sys/malloc.h>
#include <sys/timeout.h>
#include <sys/signalvar.h>
#include <sys/syslog.h>

#include "raceprot.h"
#include "lock.h"


MOD_MISC("raceprot")


int 		verbose = 0;
int 		suspend = 1;	/* suspend processes instead of killing them */
int		deny_all = 1;	/* 1 - deny_all, 0 - permit_all policy */


int		unloading_module, loadavg, timeout_purge, purge_needed;

struct timeout	to;

struct rp_entry	free_list = {NULL, 0, 0, 0, 0, 0, NONE};
int		free_list_num = 0;



#define	rp_HASH_SIZE		997
struct rp_entry	*rp_hash[rp_HASH_SIZE];

#define	stat_HASH_SIZE		7
struct rp_entry	*stat_hash[stat_HASH_SIZE];

#define	unlink_HASH_SIZE	7
struct rp_entry	*unlink_hash[unlink_HASH_SIZE];


#define	HASH_FIRST(name, value)	(name##_hash[value % name##_HASH_SIZE])





/*
 * rp_process:	Verify that the current operation does not interfere with
 *		any previous operations involving the same inode
 *
 * input:	pid of the process invoking the operation
 *		inode involved in the operation
 *		id of a filesystem where inode is residing
 *		[file]name associated with the inode
 *		operation being carried on the inode
 *		process_special_cases is set to 1 if 2 special OPEN_CREAT cases
 *			need to be checked (used only by raceprot_sys_open)
 *
 * output:	RP_PROCESS_OK		system call should be allowed to proceed
 *		RP_PROCESS_INSERT	system call should be allowed to proceed
 *					and information about the operation
 *					should be saved (rp_insert())
 *		RP_PROCESS_RESPOND	process should be suspended/killed
 *		RP_PROCESS_FAIL_CALL	system call should be aborted
 */

int
rp_process(pid_t pid, int inode, int fsid, char *name, enum FILEOP operation,
	   int process_special_cases)
{
	struct rp_entry	*tmp;
	int		found = 0;


	/* purge_needed flag is set by the callout procedure every second */

	if (purge_needed)
	{
		/*
		 * Dynamically calculate the timeout used to determine when
		 * to purge operation entries. The higher system load average
		 * is, the longer we keep the entries.
		 */

		loadavg = (averunnable.ldavg[0] * 100 + FSCALE / 2) >> FSHIFT;

		timeout_purge = min(TIMEOUT_PURGE + loadavg, TIMEOUT_PURGE_MAX);

		rp_purge(PURGE_TYPE_TIMEOUT, 0, 0);
		purge_needed = 0;
	}


	if (lock_shared() != 0)
	{
		uprintf("rp_process: failed to obtain a shared lock\n");
		return(0);
	}


	/* deal with 2 special cases */

	if (process_special_cases && (operation & OPEN_CREAT))
	{
		/*
		 * special case 1:
	 	 *
		 * temporary file race / [l]stat/access followed by open:
		 *	Victim:		calls [l]stat/access on a file &
		 *			file does not exist
		 *	Attacker:	creates a link with the same name
		 *	Victim:		open(O_CREAT, name) opens the existing
		 * 			file and trashes its contents
		 *
		 * This is a special case since stat/access is called on a file
		 * that does not exist (no inode value to process). The race
		 * condition is handled by failing the open call in step 3.
		 * It is impossible to stop the race condiiton in step 2 since
		 * the name used by an attacker won't necessarily match the
		 * name used in step 1 (i.e. /tmp/race vs /tmp/../tmp/race)
		 * and there are no inodes values to compare.
		 */

		/* only failed stat, lstat & access call entries are placed on
			stat_hash */

		for (tmp = HASH_FIRST(stat, pid); tmp != NULL; tmp = tmp->next)
		{
			if (pid == tmp->pid && strcmp(tmp->name, name) == 0)
			{
				lock_release();
				return(RP_PROCESS_FAIL_CALL);
			}
		}


		/*
		 * special case 2:
		 *
		 * unlink followed by a write:
		 *	Victim:		deletes a file
		 *	Attacker:	creates a link with the same name
		 *	Victim:		open(O_CREAT, name) opens the existing
		 * 			file and trashes its contents
		 *
		 *
		 * This is a special case since write/creat open call is called
		 * on a file that we don't know the inode value of. It is
		 * impossible to stop the race condiiton in step 2 since the
		 * name used by an attacker won't necessarily match the name
		 * used in step 1 (i.e. /tmp/race vs /tmp/../tmp/race) and
		 * there are no inode values to compare.
		 */

		/* only recently deleted file names are placed on unlink_hash */

		for (tmp = HASH_FIRST(unlink, pid); tmp != NULL; tmp= tmp->next)
		{
			if (pid == tmp->pid && strcmp(tmp->name, name) == 0)
			{
				lock_release();
				return(RP_PROCESS_FAIL_CALL);
			}
		}
	}


	/*
	 * Normal case
	 *
	 * check the operation on the current inode against all the previous
	 * operations involving the same inode
	 */

	for (tmp = HASH_FIRST(rp, inode); tmp != NULL; tmp = tmp->next)
	{
		if (tmp->inode != inode || tmp->fsid != fsid)
			continue;

		found = 1;


		if (tmp->pid == pid)
		{
			/* replace the current entry with the new operation */

			tmp->operation = operation;
			tmp->tick = ticks;

			/* we have checked the new operation against all
			   the preceeding operations until the current one
			   (they are sorted by inserted time). Since no
			   violations were detected until this point, we can
			   safely exit.
			*/

			lock_release();
			return(RP_PROCESS_OK);
		}


		if (tmp->operation == UNLINK)
		{
			/*
			 * inode collision:
			 * since the previous operation is unlink,
			 * the only way we can reach this code is if the
			 * inode is reused by another file
			 */

			//continue;
			//uprintf("inode collision %d %s\n", tmp->inode, tmp->name);

			lock_release();
			return(RP_PROCESS_OK);
		}


		if (deny_all)
		{
			if (deny_all_policy(tmp->operation, operation) ==
								POLICY_OK)
			{
				lock_release();
				return(RP_PROCESS_OK);
			}
		}
		else
		{
			if (permit_all_policy(tmp->operation, operation) ==
								POLICY_NOT_OK)
			{
				lock_release();
				return(RP_PROCESS_RESPOND);
			}
		}

		if (deny_all)
			log(9, "raceprot: Unknown operation sequence: original_pid=%d current_pid=%d inode=%d (%s) original_operation=%s current_operation=%s\n", tmp->pid, pid, inode, tmp->name, fileop_name(tmp->operation), fileop_name(operation));
	}



	/* found at least one matching inode entry, not our pid though.
	   possible interference. */

	if (found)
	{
		lock_release();

		/* deny all policy denies all requests by default */
		if (deny_all)
			return(RP_PROCESS_RESPOND);
		else
			return(RP_PROCESS_OK);
	}


	lock_release();

	return(RP_PROCESS_INSERT);
}



/*
 * deny_all_policy:	Deny All Ruleset Implementation
 *
 *			make sure the current operation doesn't interfere with
 *			any ongoing operations involving the same inode.
 *			reject any sequences of operations that are not known
 *			to be valid.
 *
 * input:		previous_op	previously observed operation involving
 *					the same inode as current_op operation
 *			current_op	current inode operation
 *
 * output:		POLICY_OK	no race conditions found
 *			POLICY_NOT_OK	race condition found
 */

int
deny_all_policy(int previous_op, int current_op)
{

#include "deny.policy"

	return(POLICY_NOT_OK);
}



/*
 * permit_all_policy:	Permit All Ruleset Implementation
 *
 *			make sure the current operation doesn't interfere with
 *			any ongoing operations involving the same inode.
 *			reject any sequences of operations that are known to
 *			be bad.
 *
 * input:		previous_op	previously observed operation involving
 *					the same inode as current_op operation
 *			current_op	current inode operation
 *
 * output:		POLICY_OK	no race conditions found
 *			POLICY_NOT_OK	race condition found
 */

int
permit_all_policy(int previous_op, int current_op)
{

#include "permit.policy"

	return(POLICY_OK);
}



/*
 * rp_register_fd:	register a file descriptor with an existing entry
 *			(only used by sys_open)
 *
 * input:		pid & inode of the current process & operation
 *			fd file descriptor to register
 *
 * output:		1 in case of success
 *			0 in case of failure (lock failed, no matching entry found)
 */

#ifdef MEDIATE_SYS_CLOSE

int
rp_register_fd(pid_t pid, int inode, int fd)
{
	struct rp_entry	*tmp;


	if (lock_shared() != 0)
	{
		uprintf("rp_register_fd(): failed to obtain a shared lock\n");
		return(0);
	}


	for (tmp = HASH_FIRST(rp, inode); tmp != NULL; tmp = tmp->next)
	{
		if (tmp->inode == inode && tmp->pid == pid && tmp->fsid == fsid)
		{
			/* only open operations can have a file descriptor
				associated with them */

			if (tmp->operation & OPEN)
			{
				tmp->fd = fd;
			}

			lock_release();

			return(1);
		}
	}

	lock_release();

	return(0);
}

#endif


/*
 * rp_insert_nolock:	creates a new entry
 *
 * input:	pid	pid of the process invoking the operation
 *		inode	inode involved in the operation
 *		fsid	.
 *		name	[file]name associated with the inode
 *		operation operation being carried on the inode
 *		op_tick	current timestamp
 *		op_type	operation type:
 *			RP_OP_NORMAL	all entries
 *			RP_OP_STAT	failed stat()/lstat()/access() entries
 *			RP_OP_UNLINK	unlink()'ed entries
 *
 * output:	1 in case of success
 *		0 in case of failure (out of memory)
 *
 * Note: rp_insert_nolock() expects exclusive lock to be acquired already
 */

int
rp_insert_nolock(pid_t pid, int inode, int fsid, char *name, enum FILEOP operation, int op_tick, int op_type)
{
	struct rp_entry	*new, *tmp;


	if (op_type == RP_OP_NORMAL)
	{
		/* find the last element */

		for (tmp = HASH_FIRST(rp, inode);
			tmp != NULL && tmp->next != NULL; tmp = tmp->next)
		    ;
	}
	else if (op_type == RP_OP_STAT)
	{
		for (tmp = HASH_FIRST(stat, pid);
			tmp != NULL && tmp->next != NULL; tmp = tmp->next)
		{
			if (tmp->pid == pid && strcmp(tmp->name, name) == 0)
			{
				/* do not insert duplicates */
				return(1);
			}
		}
	}
	else if (op_type == RP_OP_UNLINK)
	{
		for (tmp = HASH_FIRST(unlink, pid);
			tmp != NULL && tmp->next != NULL; tmp = tmp->next)
		{
			if (tmp->pid == pid && strcmp(tmp->name, name) == 0)
			{
				/* do not insert duplicates */
				return(1);
			}
		}
	}
	else
	{
		uprintf("rp_insert: Unknown operation type\n");
		return(0);
	}


	new = rp_malloc(strlen(name)+1);

	if (new == NULL)
		return(ENOMEM);

	strlcpy(new->name, name, strlen(name) + 1);

	new->next = NULL;
	new->tick = op_tick;
	new->pid = pid;
	new->inode = inode;
	new->fsid = fsid;
	new->operation = operation;
	new->fd = 0;


	if (tmp == NULL)
	{
		/* fill in the first entry of the hash chain */

		if (op_type == RP_OP_NORMAL)
		{
			HASH_FIRST(rp, inode) = new;
		}
		else if (op_type == RP_OP_STAT)
		{
			HASH_FIRST(stat, pid) = new;
		}
		else if (op_type == RP_OP_UNLINK)
		{
			HASH_FIRST(unlink, pid) = new;
		}
	}
	else
	{
		tmp->next = new;
	}

	return(1);
}



/*
 * rp_insert:		rp_insert_nolock() wrapper
 *			obtain an exclusive lock and call rp_insert_nolock()
 *
 * input/output:	see rp_insert_nolock()
 */

inline int
rp_insert(pid_t pid, int inode, int fsid, char *name, enum FILEOP operation, int op_tick, int op_type)
{
	int	ret;

	if (lock_exclusive() != 0)
	{
		uprintf("rp_insert: failed to obtain an exclusive lock\n");
		return(0);
	}

	ret = rp_insert_nolock(pid, inode, fsid, name, operation, op_tick, op_type);

	lock_release();

	return(ret);
}



/*
 * rp_malloc:	allocate the memory for the new entry (structure + filename)
 *
 * input:	size of the filename
 *
 * output:	pointer to the allocated structure
 *
 * Note:	needs exclusive access to modify free_list
 *		(provided by rp_insert lock)
 */

inline struct rp_entry*
rp_malloc(int size)
{
	struct rp_entry	*new;


	/*
	 * call malloc only when
	 * - free_list is empty or
	 * - size of the filename is over 31 characters long
	 *   (we don't cache entries with long filenames)
	 * otherwise return an entry from a free_list
	 */

	if (free_list_num == 0 || size > 31)
	{
		new = (struct rp_entry*) malloc(sizeof(struct rp_entry), M_FREE, M_NOWAIT);

		if (new == NULL)
		{
			uprintf("new = NULL in rp_malloc()\n");
			return(NULL);
		}

		new->name = (char*) malloc(size > 31 ? size : 32, M_FREE, M_NOWAIT);

		if (new->name == NULL)
		{
			uprintf("new->name = NULL in rp_malloc()\n");
			free(new, M_FREE);
			return(NULL);
		}
	}
	else
	{
		/* return the first entry off of a free_list */

		new = free_list.next;
		free_list.next = free_list.next->next;

		--free_list_num;
	}

	return(new);
}



/*
 * rp_free:	add the entry to a free_list unless the free_list is full
 *		already, in that case free() the entry
 *		
 * input:	ptr the entry to deallocate
 *
 * output:	none
 *
 * Note:	needs exclusive access to modify free_list
 *	 	(provided by rp_purge() lock)
 */

inline void
rp_free(struct rp_entry *ptr)
{
	/*
	 * do not cache the entries if
	 *	- free list is full or
	 *	- size of the string is over 31 bytes or
	 *	- we are unloading the module
	 */

	if ((free_list_num > FREE_LIST_NUM) || strlen(ptr->name) > 31 || unloading_module)
	{
		free(ptr->name, M_FREE);
		free(ptr, M_FREE);
	}
	else
	{
		/* append ptr to the free_list */

		ptr->next = free_list.next;
		free_list.next = ptr;

		++free_list_num;
	}
}



/*
 * rp_purge:	remove the stale (timed out) entries or
 *		all the entries if we are unloading the module
 *
 * input:	type	purge operation type
 *			PURGE_TYPE_TIMEOUT	timeout
 *			PURGE_TYPE_PID		purge by pid
 *			PURGE_TYPE_FD		purge by fd
 *
 *		pid	pid of the entries to remove
 *		(pid is set to 0 when we are removing in case of a timeout)
 *
 *		fd file descriptor of the entries to remove
 *		(fd is set to 0 when we are removing by pid or a timeout)
 *
 * output:	none
 */


void
rp_purge(int type, int pid, int fd)
{
	struct rp_entry	*ptr, *next, *prev, *tmp;
	int		i;


	if (lock_exclusive() != 0)
	{
		uprintf("rp_purge: failed to obtain an exclusive lock\n");
		return;
	}


#define	RP_FREE()					\
	tmp = ptr;					\
	if (prev != NULL)				\
		prev->next = next;			\
	else						\
		rp_hash[i] = ptr = next;		\
	rp_free(tmp);


	for (i = 0; i < rp_HASH_SIZE; i++)
	{
		prev = NULL;
		ptr = rp_hash[i];

		for (; ptr != NULL; )
		{
			next = ptr->next;

			if (type & PURGE_TYPE_TIMEOUT)
			{
				if (((ticks - ptr->tick) > timeout_purge) ||
				    unloading_module)
				{
					RP_FREE();
				}
				else
				{
					/*
					 * if purging by timeout, we can bail
					 * out earlier by giving up on the first
					 * entry that fails the time comparison
					 * since the entries are sorted by
					 * inserted time
					 */

					break;
				}
			}
#ifdef MEDIATE_SYS_CLOSE
			else if ((type & PURGE_TYPE_FD) &&
					ptr->pid == pid && ptr->fd == fd)
			{
				/* if we are purging by a file descriptor
				   then bail after the first match since there
				   can by only one unique descriptor
				   associated with each pid */

				RP_FREE();

				lock_release();
				return;
			}
#endif
			else if (((type & PURGE_TYPE_PID) && ptr->pid == pid))
			{
				RP_FREE();
			}
			else
			{
				prev = ptr;
			}

			ptr = next;
		}
	}


	/* clean the stat hash */

	for (i = 0; i < stat_HASH_SIZE; i++)
	{
		prev = NULL;
		ptr = stat_hash[i];

		for (; ptr != NULL; )
		{
			next = ptr->next;

			if ((ticks - ptr->tick) > timeout_purge)
			{
				tmp = ptr;	

				if (prev != NULL)
					prev->next = next;
				else
					stat_hash[i] = ptr = next;
			
				rp_free(tmp);
			}
			else
			{
				prev = ptr;
			}

			ptr = next;
		}
	}


	/* clean the unlink hash */

	for (i = 0; i < unlink_HASH_SIZE; i++)
	{
		prev = NULL;
		ptr = unlink_hash[i];

		for (; ptr != NULL; )
		{
			next = ptr->next;

			if ((ticks - ptr->tick) > timeout_purge)
			{
				tmp = ptr;	

				if (prev != NULL)
					prev->next = next;
				else
					unlink_hash[i] = ptr = next;
			
				rp_free(tmp);
			}
			else
			{
				prev = ptr;
			}

			ptr = next;
		}
	}


	lock_release();
}



/*
 * rp_duplicate_entry:	duplicate all the entries associated with a pid
 *			(called by sys_*fork)
 *
 * input:		oldpid pid of the entries to duplicate
 *			newpid new pid to associate all entries with
 *
 * output:		none
 */

void
rp_duplicate_entry(int oldpid, int newpid)
{
	struct rp_entry	*ptr;
	int		i;


	if (lock_exclusive() != 0)
	{
		uprintf("rp_duplicate_entry: failed to obtain an exclusive lock\n");
		return;
	}


	for (i = 0; i < rp_HASH_SIZE; i++)
	{
		for (ptr = rp_hash[i]; ptr != NULL; ptr = ptr->next)
		{
			if (ptr->pid == oldpid)
			{
				rp_insert_nolock(newpid, ptr->inode, ptr->fsid,
						 ptr->name, ptr->operation,
						 ticks, RP_OP_NORMAL);
			}
		}
	}


	lock_release();
}



/*
 * rp_timeout:	soft timer callout (called every TIMEOUT_TIMER/1000 seconds)
 *		call rp_purge to purge the stale entries
 *
 * input:	arg timeout_add parameter
 *
 * output:	none
 */

void
rp_timeout(void *arg)
{
	purge_needed = 1;

	timeout_add((struct timeout*)arg, TIMEOUT_TIMER);
}



/*
 * rp_init:	Initialization function called upon module loading.
 *		Initializes the locks, timeouts and various data structures
 *
 * input:	none
 *
 * output:	none
 */

void
rp_init()
{
	bzero(&rp_hash, sizeof(rp_hash));
	bzero(&stat_hash, sizeof(stat_hash));
	bzero(&unlink_hash, sizeof(unlink_hash));

	unloading_module = 0;

	timeout_set(&to, rp_timeout, &to);
	timeout_add(&to, TIMEOUT_TIMER);

	lockinit(&rp_lock, PCATCH, "tm", 0, 0);
}



/*
 * rp_fini:	Called upon module unloading.
 *		Removes the timeout & frees the data structures
 *
 * input:	none
 *
 * output:	none
 */

void
rp_fini()
{
	struct rp_entry	*ptr, *tmp;
	int		i;


	unloading_module = 1;

	timeout_del(&to);

	rp_purge(PURGE_TYPE_TIMEOUT, 0, 0);


	/* free all the entries on the free_list */

	for (ptr = free_list.next; ptr != NULL; )
	{
		tmp = ptr;
		ptr = ptr->next;

		free(tmp->name, M_FREE);
		free(tmp, M_FREE);
	}


	/* free all the stat hash entries */

	for (i = 0; i < stat_HASH_SIZE; i++)
	{
		for (ptr = stat_hash[i]; ptr != NULL; )
		{
			tmp = ptr;
			ptr = ptr->next;

			free(tmp->name, M_FREE);
			free(tmp, M_FREE);
		}
	}


	/* free all the unlink hash entries */

	for (i = 0; i < unlink_HASH_SIZE; i++)
	{
		for (ptr = unlink_hash[i]; ptr != NULL; )
		{
			tmp = ptr;
			ptr = ptr->next;

			free(tmp->name, M_FREE);
			free(tmp, M_FREE);
		}
	}
}


int raceprot_sys_exit(struct proc*, void*, register_t*);
int raceprot_sys_fork(struct proc*, void*, register_t*);
int raceprot_sys_rfork(struct proc*, void*, register_t*);
int raceprot_sys_vfork(struct proc*, void*, register_t*);
int raceprot_sys_execve(struct proc*, void*, register_t*);
int raceprot_sys_open(struct proc*, void*, register_t*);
int raceprot_sys_link(struct proc*, void*, register_t*);
int raceprot_sys_unlink(struct proc*, void*, register_t*);
int raceprot_sys_chdir(struct proc*, void*, register_t*);
int raceprot_sys_mknod(struct proc*, void*, register_t*);
int raceprot_sys_chmod(struct proc*, void*, register_t*);
int raceprot_sys_chown(struct proc*, void*, register_t*);
int raceprot_sys_access(struct proc*, void*, register_t*);
int raceprot_sys_symlink(struct proc*, void*, register_t*);
int raceprot_sys_readlink(struct proc*, void*, register_t*);
int raceprot_sys_rename(struct proc*, void*, register_t*);
int raceprot_sys_mkfifo(struct proc*, void*, register_t*);
int raceprot_sys_mkdir(struct proc*, void*, register_t*);
int raceprot_sys_rmdir(struct proc*, void*, register_t*);
int raceprot_sys_utimes(struct proc*, void*, register_t*);
int raceprot_sys_stat(struct proc*, void*, register_t*);
int raceprot_sys_lstat(struct proc*, void*, register_t*);
int raceprot_sys_truncate(struct proc*, void*, register_t*);
int raceprot_sys_lchown(struct proc*, void*, register_t*);


struct backup_sysent backup_sysent[] = {
#define NUM_SYS_exit 0
	BACKUP_SYSENT(1, exit)
#define NUM_SYS_fork 1
	BACKUP_SYSENT(2, fork)
#define NUM_SYS_rfork 2
	BACKUP_SYSENT(251, rfork)
#define NUM_SYS_vfork 3
	BACKUP_SYSENT(66, vfork)
#define NUM_SYS_execve 4
	BACKUP_SYSENT(59, execve)
#define NUM_SYS_open 5
	BACKUP_SYSENT(5, open)
#define NUM_SYS_link 6
	BACKUP_SYSENT(9, link)
#define NUM_SYS_unlink 7
	BACKUP_SYSENT(10, unlink)
#define NUM_SYS_chdir 8
	BACKUP_SYSENT(12, chdir)
#define NUM_SYS_mknod 9
	BACKUP_SYSENT(14, mknod)
#define NUM_SYS_chmod 10
	BACKUP_SYSENT(15, chmod)
#define NUM_SYS_chown 11
	BACKUP_SYSENT(16, chown)
#define NUM_SYS_access 12
	BACKUP_SYSENT(33, access)
#define NUM_SYS_symlink 13
	BACKUP_SYSENT(57, symlink)
#define NUM_SYS_readlink 14
	BACKUP_SYSENT(58, readlink)
#define NUM_SYS_rename 15
	BACKUP_SYSENT(128, rename)
#define NUM_SYS_mkfifo 16
	BACKUP_SYSENT(132, mkfifo)
#define NUM_SYS_mkdir 17
	BACKUP_SYSENT(136, mkdir)
#define NUM_SYS_rmdir 18
	BACKUP_SYSENT(137, rmdir)
#define NUM_SYS_utimes 19
	BACKUP_SYSENT(138, utimes)
#define NUM_SYS_stat 20
	BACKUP_SYSENT(188, stat)
#define NUM_SYS_lstat 21
	BACKUP_SYSENT(190, lstat)
#define NUM_SYS_truncate 22
	BACKUP_SYSENT(200, truncate)
#define NUM_SYS_lchown 23
	BACKUP_SYSENT(254, lchown)
};




char*
fileop_name(enum FILEOP operation)
{
	if (operation & OPEN_READ)		return "ropen";
	if (operation & OPEN_WRITE)	return "wopen";
	if (operation & OPEN_CREAT)	return "copen";
	if (operation & RENAME_FROM)	return "frename";
	if (operation & RENAME_TO)		return "trename";
	if (operation & STAT)		return "stat";
	if (operation & LSTAT)		return "lstat";
	if (operation & CLOSE)		return "close";
	if (operation & EXEC)		return "exec";
	if (operation & LINK)		return "link";
	if (operation & UNLINK)		return "unlink";
	if (operation & CHDIR)		return "chdir";
	if (operation & MKNOD)		return "mknod";
	if (operation & CHMOD)		return "chmod";
	if (operation & CHOWN)		return "chown";
	if (operation & ACCESS)		return "access";
	if (operation & SYMLINK)		return "symlink";
	if (operation & READLINK)		return "readlink";
	if (operation & MKFIFO)		return "mkfifo";
	if (operation & MKDIR)		return "mkdir";
	if (operation & RMDIR)		return "rmdir";
	if (operation & UTIMES)		return "utimes";
	if (operation & TRUNCATE)		return "truncate";
	if (operation & LCHOWN)		return "lchown";

	return "none";
}

/*
 * Mediate sys_link system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_link(struct proc *p, void *v, register_t *retval)
{
	struct sys_link_args /* {
		syscallarg(char *) path2;
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_link(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(link);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), LINK, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, LINK, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(link);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), LINK, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_unlink system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_unlink(struct proc *p, void *v, register_t *retval)
{
	struct sys_unlink_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_unlink(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(unlink);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), UNLINK, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, UNLINK, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(unlink);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), UNLINK, ticks, RP_OP_UNLINK);
	}

	return ret2;
}



/*
 * Mediate sys_chdir system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_chdir(struct proc *p, void *v, register_t *retval)
{
	struct sys_chdir_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_chdir(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(chdir);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), CHDIR, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, CHDIR, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(chdir);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), CHDIR, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_mknod system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_mknod(struct proc *p, void *v, register_t *retval)
{
	struct sys_mknod_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_mknod(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(mknod);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), MKNOD, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, MKNOD, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(mknod);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), MKNOD, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_chmod system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_chmod(struct proc *p, void *v, register_t *retval)
{
	struct sys_chmod_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_chmod(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(chmod);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), CHMOD, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, CHMOD, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(chmod);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), CHMOD, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_chown system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_chown(struct proc *p, void *v, register_t *retval)
{
	struct sys_chown_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_chown(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(chown);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), CHOWN, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, CHOWN, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(chown);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), CHOWN, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_access system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_access(struct proc *p, void *v, register_t *retval)
{
	struct sys_access_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_access(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		/* save the information about the non-existent file */
		rp_insert(p->p_pid, 0, 0, (char*) SCARG(uap, path), ACCESS, ticks, RP_OP_STAT);

		return CALL_SAVED_SYS_CALL(access);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), ACCESS, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, ACCESS, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(access);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), ACCESS, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_symlink system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_symlink(struct proc *p, void *v, register_t *retval)
{
	struct sys_symlink_args /* {
		syscallarg(char *) path2;
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_symlink(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(symlink);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), SYMLINK, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, SYMLINK, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(symlink);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), SYMLINK, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_readlink system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_readlink(struct proc *p, void *v, register_t *retval)
{
	struct sys_readlink_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_readlink(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(readlink);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), READLINK, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, READLINK, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(readlink);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), READLINK, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_mkfifo system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_mkfifo(struct proc *p, void *v, register_t *retval)
{
	struct sys_mkfifo_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_mkfifo(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(mkfifo);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), MKFIFO, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, MKFIFO, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(mkfifo);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), MKFIFO, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_mkdir system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_mkdir(struct proc *p, void *v, register_t *retval)
{
	struct sys_mkdir_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_mkdir(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(mkdir);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), MKDIR, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, MKDIR, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(mkdir);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), MKDIR, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_rmdir system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_rmdir(struct proc *p, void *v, register_t *retval)
{
	struct sys_rmdir_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_rmdir(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(rmdir);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), RMDIR, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, RMDIR, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(rmdir);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), RMDIR, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_utimes system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_utimes(struct proc *p, void *v, register_t *retval)
{
	struct sys_utimes_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_utimes(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(utimes);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), UTIMES, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, UTIMES, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(utimes);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), UTIMES, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_truncate system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_truncate(struct proc *p, void *v, register_t *retval)
{
	struct sys_truncate_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_truncate(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(truncate);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), TRUNCATE, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, TRUNCATE, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(truncate);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), TRUNCATE, ticks, RP_OP_NORMAL);
	}

	return ret2;
}



/*
 * Mediate sys_lchown system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_lchown(struct proc *p, void *v, register_t *retval)
{
	struct sys_lchown_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d sys_lchown(%s)\n", p->p_pid, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file object exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(lchown);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), LCHOWN, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, LCHOWN, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	ret2 = CALL_SAVED_SYS_CALL(lchown);

	if (ret != RP_PROCESS_OK && ret2 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), LCHOWN, ticks, RP_OP_NORMAL);
	}

	return ret2;
}




/*
 * raceprot_respond:	Respond to a possible race condition attack:
 *			either suspend a process (depending on a suspend var)
 *			or kill it
 *
 * input:		struct proc *p		used for sending a kill signal
 *			enum FILEOP operation	operation that caused the
 *						race condition
 *			const char *path	name of a file object that is
 *						involved
 *			int obj_type		type (directory / file) of a
 *						file object that caused the
 *						race condition
 */

void
raceprot_respond(struct proc *p, enum FILEOP operation, const char *path, int obj_type)
{
	if (suspend)
	{
		int	timeout;

		timeout = (obj_type == VDIR ? TIMEOUT_DIR : TIMEOUT_FILE);
		timeout += loadavg;

		log(9, "Potentially dangerous operation %d %s(%s). Suspending process %d (%s) for %d seconds\n", operation, fileop_name(operation), path, p->p_pid, p->p_comm, timeout / 100);

		tsleep("susp", PZERO, "raceprot", timeout);
		//tsleep("susp", PPAUSE, "raceprot", timeout);
	}
	else
	{
		log(9, "Potentially dangerous operation %s(%s). Killing process %d (%s)\n", fileop_name(operation), path, p->p_pid, p->p_comm);

		psignal(p, 9);
	}
}



/*
 * Mediate sys_stat system call:
 *
 * Call the original stat system call.
 *
 * We save information about non-existent files
 * to detect temporary file race conditions.
 */

int
raceprot_sys_stat(struct proc *p, void *v, register_t *retval)
{
        struct sys_stat_args /* {
                syscallarg(char *) path;
        } */ *uap = v;
        int                     ret;


	ret = CALL_SAVED_SYS_CALL(stat);


	if (verbose)
		uprintf("%d sys_stat(%s) = %d\n", p->p_pid, SCARG(uap, path), ret);


	/* did stat fail? */
	if (ret == ENOENT)
	{
		/* save information about the non-existent file */

		rp_insert(p->p_pid, 0, 0, (char*) SCARG(uap, path), STAT,
			  ticks, RP_OP_STAT);
	}

	return ret;
}



/*
 * Mediate sys_lstat system call:
 *
 * Call the original stat system call.
 *
 * We save information about non-existent files
 * to detect temporary file race conditions.
 */

int
raceprot_sys_lstat(struct proc *p, void *v, register_t *retval)
{
        struct sys_lstat_args /* {
                syscallarg(char *) path;
        } */ *uap = v;
        int                     ret;


	ret = CALL_SAVED_SYS_CALL(lstat);


	if (verbose)
		uprintf("%d sys_lstat(%s) = %d\n", p->p_pid, SCARG(uap, path), ret);


	/* did lstat fail? */
	if (ret == ENOENT)
	{
		/* save information about the non-existent file */

		rp_insert(p->p_pid, 0, 0, (char*) SCARG(uap, path), LSTAT,
			  ticks, RP_OP_STAT);
	}

	return ret;
}



/*
 * raceprot_sys_open:	Mediate sys_open call
 *
 * Note: there is a race condition between this function and sys_open:
 *	 a filename can be changed in the userspace before sys_open gets to
 *	 process it (reference: LOMAC you can live with)
 */

int
raceprot_sys_open(struct proc *p, void *v, register_t *retval)
{
	struct sys_open_args /* {
		syscallarg(char *) path;
		syscallarg(int) flags;
		syscallarg(int) mode;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2, op;
	struct vattr		va;


	if (verbose)
		uprintf("%d (%s) sys_open(%s)\n", p->p_pid, p->p_comm,
							SCARG(uap, path));

	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file exist? */

	if (namei(&nd) == 0)
	{
		/* get the inode */

		VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);

		op = SCARG(uap, flags) & (O_RDWR | O_WRONLY | O_CREAT) ? OPEN_WRITE : OPEN_READ;

		/* special case, insecure O_CREAT */
		if (SCARG(uap, flags) & (O_CREAT & ~O_EXCL))
			op = op | OPEN_CREAT;


		ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
					(char*) SCARG(uap, path), op, 1);


		if (ret == RP_PROCESS_RESPOND)
		{
			raceprot_respond(p, op, SCARG(uap, path),
							nd.ni_vp->v_type);
		}
		else if (ret == RP_PROCESS_FAIL_CALL)
		{
			/* fail the insecure O_CREAT. force the O_EXCL behavior */

			if (verbose)
				uprintf("insecure O_CREAT detected\n");

			return(EEXIST);
		}


		ret2 = CALL_SAVED_SYS_CALL(open);

		if (ret != RP_PROCESS_OK && ret2 == 0)
		    rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			      (char*)SCARG(uap, path), op, ticks, RP_OP_NORMAL);


		if (verbose)
			uprintf("%d sys_open(%s, %d) = %d (fd %d)\n", p->p_pid, SCARG(uap, path), op, ret, retval[0]);


#ifdef MEDIATE_SYS_CLOSE
		if (ret2 == 0 && retval[0] != -1)
			rp_register_fd(p->p_pid, va.va_fileid, retval[0]);
#endif

		vrele(nd.ni_vp);

		return(ret2);
	}


	/* CASE 2: file does not exit. open might create it */

	ret2 = CALL_SAVED_SYS_CALL(open);

	ret = namei(&nd);

	if (verbose)
		uprintf("%d (2) sys_open(%s) = %d (fd %d)\n", p->p_pid, SCARG(uap, path), ret2, retval[0]);


	if (ret == 0)
	{
		/* get the inode */

		VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);

		op = SCARG(uap, flags) & (O_RDWR | O_WRONLY | O_CREAT) ? OPEN_WRITE : OPEN_READ;

		/* special case, insecure O_CREAT */
		if (SCARG(uap, flags) & (O_CREAT & ~O_EXCL))
			op = op | OPEN_CREAT;


		ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
					(char*) SCARG(uap, path), op, 0);


		if (ret == RP_PROCESS_RESPOND)
		{
			/* too late to respond since the system
				call has finished by now! */

			uprintf("raceprot: sys_open() finished already (%d %s, %s)\n", op, SCARG(uap, path), p->p_comm);
/*
			raceprot_respond(p, op, SCARG(uap, path), nd.ni_vp->v_type);
*/
		}
		else if (ret == RP_PROCESS_FAIL_CALL)
		{
			uprintf("raceprot: fail call sys_open() finished already (%d %s)\n", op, SCARG(uap, path));
		}


		if (ret != RP_PROCESS_OK && ret2 == 0)
		    rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			      (char*)SCARG(uap, path), op, ticks, RP_OP_NORMAL);


#ifdef MEDIATE_SYS_CLOSE
		if (retval[0] != -1)
			rp_register_fd(p->p_pid, va.va_fileid, retval[0]);
#endif

	}

	if (nd.ni_vp != NULL)
		vrele(nd.ni_vp);

	return(ret2);
}



/*
 * Mediate sys_execve system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_execve(struct proc *p, void *v, register_t *retval)
{
	struct sys_execve_args /* {
		syscallarg(char *) path;
	} */ *uap = v;
	struct nameidata	nd;
	int			ret, ret2;
	struct vattr		va;


	if (verbose)
		uprintf("%d (%s) sys_execve(%s)\n", p->p_pid, p->p_comm, SCARG(uap, path));


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, path), p);


	/* does file exist? */

	if (namei(&nd) != 0)
	{
		return CALL_SAVED_SYS_CALL(execve);
	}


	/* file object exists, retrieve its inode */

	VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


	ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
			 (char*) SCARG(uap, path), EXEC, 0);

	if (ret == RP_PROCESS_RESPOND)
	{
		raceprot_respond(p, EXEC, SCARG(uap, path), nd.ni_vp->v_type);
	}

	vrele(nd.ni_vp);


	if (ret != RP_PROCESS_OK)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
		          (char*) SCARG(uap, path), EXEC, ticks, RP_OP_NORMAL);
	}


	ret2 = CALL_SAVED_SYS_CALL(execve);

		
	/* we don't trust the exec'ed process thus we remove all the entries
	   belonging to the original process */

	if (ret2 == 0)
		rp_purge(PURGE_TYPE_PID, p->p_pid, 0);

	return(ret2);
}



/*
 * Mediate sys_rename system call:
 * Resolve pathname to an inode, check for any possible interferences,
 * suspend / kill the process if necessary
 */

int
raceprot_sys_rename(struct proc *p, void *v, register_t *retval)
{
	struct sys_rename_args /* {
		syscallarg(char *) from;
		syscallarg(char *) to;
	} */ *uap = v;
	struct nameidata	nd, nd2;
	int			ret, ret2, ret3;
	struct vattr		va;


	/* resolve pathname to an inode */

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, from), p);
	NDINIT(&nd2, LOOKUP, NOFOLLOW, UIO_USERSPACE, SCARG(uap, to), p);

	ret = namei(&nd);
	ret2 = namei(&nd2);

	if (verbose)
		uprintf("%d (%s) sys_rename(%s, %s) = %d %d\n", p->p_pid, p->p_comm, SCARG(uap, from), SCARG(uap, to), ret, ret2);


	/* does file exist? */

	if (ret == 0)
	{
		/* get the inode */

		VOP_GETATTR(nd.ni_vp, &va, p->p_ucred, p);


		ret = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
				 (char*) SCARG(uap, from), RENAME_FROM, 0);

		if (ret == RP_PROCESS_RESPOND)
		{
			raceprot_respond(p, RENAME_FROM, SCARG(uap, from),
							 nd.ni_vp->v_type);
		}

		vrele(nd.ni_vp);


		if (ret2 == 0)
		{
			/* get the inode */

			VOP_GETATTR(nd2.ni_vp, &va, p->p_ucred, p);


			ret2 = rp_process(p->p_pid, va.va_fileid, va.va_fsid,
				(char*) SCARG(uap, to), RENAME_TO, 0);

			if (ret2 == RP_PROCESS_RESPOND)
			{
				raceprot_respond(p, RENAME_TO, SCARG(uap, to), nd2.ni_vp->v_type);
			}

			vrele(nd2.ni_vp);
		}
	}


	ret3 = CALL_SAVED_SYS_CALL(rename);

	if (ret != RP_PROCESS_OK && ret2 != RP_PROCESS_OK && ret3 == 0)
	{
		rp_insert(p->p_pid, va.va_fileid, va.va_fsid,
			  (char*) SCARG(uap, to), RENAME_TO, ticks, RP_OP_NORMAL);
	}

	return ret3;
}



/*
 * Mediate sys_close system call:
 * purge the entry belonging to the current process with the file descriptor fd
 */

#ifdef MEDIATE_SYS_CLOSE

int
raceprot_sys_close(struct proc *p, void *v, register_t *retval)
{
	struct sys_close_args	*uap = v;
	int			ret;


	ret = CALL_SAVED_SYS_CALL(close);

	if (verbose)
		uprintf("%d sys_close(%d) %d\n", p->p_pid, SCARG(uap, fd), ret);

	if (ret == 0)
		rp_purge(PURGE_TYPE_FD, p->p_pid, SCARG(uap, fd));


	return(ret);
}

#endif



/*
 * Mediate sys_fork system call:
 * duplicate all the entries belonging to the current process with the new pid
 */

int
raceprot_sys_fork(struct proc *p, void *v, register_t *retval)
{
	int			ret;


	ret = CALL_SAVED_SYS_CALL(fork);

	if (verbose)
		uprintf("%d sys_fork() %d\n", p->p_pid, retval[0]);

	if (ret == 0)
		rp_duplicate_entry(p->p_pid, retval[0]);


	return(ret);
}


/*
 * Mediate sys_vfork system call:
 * duplicate all the entries belonging to the current process with the new pid
 */

int
raceprot_sys_vfork(struct proc *p, void *v, register_t *retval)
{
	int			ret;


	ret = CALL_SAVED_SYS_CALL(vfork);

	if (verbose)
		uprintf("%d sys_vfork() %d\n", p->p_pid, retval[0]);

	if (ret == 0)
		rp_duplicate_entry(p->p_pid, retval[0]);


	return(ret);
}



/*
 * Mediate sys_rfork system call:
 * duplicate all the entries belonging to the current process with the new pid
 */

int
raceprot_sys_rfork(struct proc *p, void *v, register_t *retval)
{
	int			ret;


	ret = CALL_SAVED_SYS_CALL(rfork);

	if (verbose)
		uprintf("%d sys_rfork() %d\n", p->p_pid, retval[0]);

	if (ret == 0)
		rp_duplicate_entry(p->p_pid, retval[0]);


	return(ret);
}



/*
 * Mediate sys_exit system call:
 * purge all the entries belonging to the terminated process
 */

int
raceprot_sys_exit(struct proc *p, void *v, register_t *retval)
{
	if (verbose)
		uprintf("%d sys_exit()\n", p->p_pid);

	rp_purge(PURGE_TYPE_PID, p->p_pid, 0);

	return CALL_SAVED_SYS_CALL(exit);
}



/*
 * Kernel Module Handler
 * Load / Unload the module
 */

static int
raceprot_handle(struct lkm_table *lkmt, int cmd)
{
	int		i, num;


	switch(cmd) {
	case LKM_E_LOAD:
		/* don't load twice! */

		if (lkmexists(lkmt))
			return (EEXIST);


		/* initialize the race protection system */

		rp_init();


		/* intercept the system calls listed in backup_sysent table */

		for (i = 0; i < BACKUP_SYSENT_NUM_ENTRIES; i++)
		{
			num = backup_sysent[i].sys_number;

			if (verbose)
				uprintf("loading %s(%d)\n",
					backup_sysent[i].sys_name, num);

			backup_sysent[i].old_sys_call = sysent[num].sy_call;
			sysent[num].sy_call = backup_sysent[i].new_sys_call;
		}

		break;


	case LKM_E_UNLOAD:

		/* restore the system calls listed in backup_sysent table */

		for (i = 0; i < BACKUP_SYSENT_NUM_ENTRIES; i++)
		{
			num = backup_sysent[i].sys_number;

			if (verbose)
				uprintf("unloading %s(%d)\n",
					backup_sysent[i].sys_name, num);

			sysent[num].sy_call = backup_sysent[i].old_sys_call;
		}

		rp_fini();


		break;
	}

	return(0);
}



/*
 * Kernel Module Entry Point
 */

int
raceprot(struct lkm_table *lkmt, int cmd, int ver)
{
	DISPATCH(lkmt, cmd, ver, raceprot_handle, raceprot_handle, lkm_nofunc)
}


